use std::time::Duration;

use anyhow::{anyhow, Context};
use config::Config;
use log::{debug, error, info};
use rand::{
    distributions::{Distribution, Standard, WeightedIndex},
    seq::IteratorRandom,
    Rng,
};
use shows::Shows;

mod config;
mod media;
mod shows;

#[tokio::main]
async fn main() -> anyhow::Result<()> {
    dotenvy::dotenv().ok();
    env_logger::init();
    ffmpeg_next::init().unwrap();

    let mut rng = rand::thread_rng();

    let conf = config::load()?;

    info!("Loading shows from {}", conf.shows_file.display());
    let shows = shows::load(&conf.shows_file).with_context(|| {
        format!(
            "Failed to load shows from file {}",
            conf.shows_file.display()
        )
    })?;
    if shows.is_empty() {
        return Err(anyhow!("Shows file is empty!"));
    }

    let dist = WeightedIndex::new(shows.iter().map(|s| s.weight))
        .context("Failed to load show weights")?;

    info!("Logging into cohost as {}", conf.cohost.email);
    let session = eggbug::Session::login(&conf.cohost.email, &conf.cohost.password)
        .await
        .context("Failed to login to cohost")?;

    let mut retry_count = 0;

    loop {
        let result = post_random_capture(&conf, &shows, &session, &dist, &mut rng)
            .await
            .context("Failed to post a random capture");

        match result {
            Ok(()) => {
                if conf.post_interval == Duration::ZERO {
                    return result;
                }
                retry_count = 0;
                tokio::time::sleep(conf.post_interval).await;
            }
            Err(ref e) => {
                error!("{:#}", e);
                if conf.max_retries.map(|m| retry_count < m) == Some(false) {
                    return result.context("Retry limit exceeded");
                }
                info!("retrying in {} seconds", conf.retry_delay.as_secs());
                retry_count += 1;
                tokio::time::sleep(conf.retry_delay).await;
            }
        }
    }
}

async fn post_random_capture<R: Rng>(
    conf: &Config,
    shows: &Shows,
    session: &eggbug::Session,
    dist: &WeightedIndex<f32>,
    rng: &mut R,
) -> anyhow::Result<()> {
    let show = &shows[dist.sample(rng)];
    let episodes = show.episodes().with_context(|| {
        format!(
            "Failed to get episode list for show {} with path {}",
            show.title,
            show.path.display()
        )
    })?;
    let (num, file) = episodes.iter().choose(rng).unwrap();

    let descriptor = shows::display_show_episode(show, num);

    info!("Selected: {} - {}", descriptor, file.display());

    let media_info = media::get_media_info(file, &conf.capture)
        .with_context(|| format!("Failed to get info for media file {}", file.display()))?;
    debug!(
        "Media duration: {}",
        format_timestamp(media_info.duration_secs, None, true)
    );
    debug!(
        "Subtitle stream index: {:?}",
        media_info.subtitle_stream_index
    );

    let max_timestamp = match conf.capture.audio {
        Some(d) => media_info.duration_secs - d,
        None => media_info.duration_secs,
    };
    if max_timestamp < 0.0 {
        return Err(anyhow!(
            "Media file {} is too short to take audio clip!",
            file.display()
        ));
    }
    let timestamp = max_timestamp * rng.sample::<f64, _>(Standard);
    info!(
        "Taking capture at {}",
        format_timestamp(timestamp, Some(media_info.duration_secs), true)
    );

    let mut attachments = Vec::new();
    let display_timestamp = format_timestamp(timestamp, Some(media_info.duration_secs), false);

    if conf.capture.images {
        let image_data = media::take_screencap(file, timestamp, media_info.subtitle_stream_index)
            .await
            .context("Failed to take screencap")?;

        let image_size = imagesize::blob_size(&image_data)
            .context("Failed to get image size for screencap image data")?;

        let image_attachment = eggbug::Attachment::new(
            image_data,
            format!("{} @ {}.png", descriptor, display_timestamp),
            String::from("image/png"),
            eggbug::MediaMetadata::Image {
                width: Some(image_size.width as u32),
                height: Some(image_size.height as u32),
            },
        )
        .with_alt_text(format!(
            "Screencap of {} at {}",
            descriptor, display_timestamp
        ));

        attachments.push(image_attachment);
    }

    if let Some(duration) = conf.capture.audio {
        let audio_data =
            media::take_audio_clip(file, timestamp, duration, media_info.audio_stream_index)
                .await
                .context("Failed to take audio clip")?;

        let audio_attachment = eggbug::Attachment::new(
            audio_data,
            format!("{} @ {}.mp3", descriptor, display_timestamp),
            String::from("audio/mpeg"),
            eggbug::MediaMetadata::Audio {
                artist: show.title.clone(),
                title: format!("{} @ {}", descriptor, display_timestamp),
            },
        );

        attachments.push(audio_attachment);
    }

    let mut tags = show.tags.clone();
    tags.extend_from_slice(&conf.metadata.global_tags);

    let chost_id = session
        .create_post(
            &conf.cohost.page,
            &mut eggbug::Post {
                content_warnings: if conf.metadata.cw {
                    vec![descriptor]
                } else {
                    vec![]
                },
                attachments,
                tags,
                draft: conf.metadata.draft,
                adult_content: show.eighteen_plus.unwrap_or(conf.metadata.eighteen_plus),
                headline: String::new(),
                markdown: String::new(),
                metadata: None,
                ask: None,
            },
        )
        .await
        .context("Failed to create chost")?;
    info!("Created post {}", chost_id);

    Ok(())
}

fn format_timestamp(
    timestamp: f64,
    total_duration: Option<f64>,
    fractional_seconds: bool,
) -> String {
    let total_duration = total_duration.unwrap_or(timestamp);

    let truncated_timestamp = timestamp as u32;

    let optional_hours = if total_duration >= 3600.0 {
        format!("{}:", truncated_timestamp / 3600)
    } else {
        String::new()
    };

    let minutes = format!("{:02}", (truncated_timestamp % 3600) / 60);

    let seconds = if fractional_seconds {
        format!("{:05.2}", timestamp % 60.0)
    } else {
        format!("{:02}", truncated_timestamp % 60)
    };

    format!("{optional_hours}{minutes}:{seconds}")
}
